-----------------------------------------------------------------------------------------------
-- Client Lua Script for ShortCircuit
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- ShortCircuit Module Definition
-----------------------------------------------------------------------------------------------
local ShortCircuit = {} 

local tSavedConfigurations = {}

local tTradeskillContainer = nil

local wndDialogStore = nil
local wndDialogYesNo = nil
local wndSelectedGridItem = nil
local wndDialogDuplicate = nil

local tSelectedCraft = nil

local iGridRow = nil

local timerCraftingStationCheck = nil
local bLastAtCraftingStationStatus = nil

local strSort = nil
local bReverseSort = false
local wndLastSort = nil

local kTradeskills =
{
	[CraftingLib.CodeEnumTradeskill.Survivalist]	=	{strName = "Survivalist",	strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Survivalist"},
	[CraftingLib.CodeEnumTradeskill.Architect]		=	{strName = "Architect",		strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Architect"},
	[CraftingLib.CodeEnumTradeskill.Fishing]		=	{strName = "Fishing",		strIcon = ""},
	[CraftingLib.CodeEnumTradeskill.Mining]			=	{strName = "Mining",		strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Miner"},
	[CraftingLib.CodeEnumTradeskill.Relic_Hunter]	=	{strName = "Relic Hunter",	strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_RelicHunter"},
	[CraftingLib.CodeEnumTradeskill.Cooking]		=	{strName = "Cooking",		strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Cooking"},
	[CraftingLib.CodeEnumTradeskill.Outfitter]		=	{strName = "Outfitter",		strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Outfitter"},
	[CraftingLib.CodeEnumTradeskill.Armorer]		=	{strName = "Armorer",		strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Armorer"},
	[CraftingLib.CodeEnumTradeskill.Farmer]			=	{strName = "Farmer",		strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Farmer"},
	[CraftingLib.CodeEnumTradeskill.Weaponsmith]	=	{strName = "Weaponsmith",	strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_WeaponCrafting"},
	[CraftingLib.CodeEnumTradeskill.Tailor]			=	{strName = "Tailor",		strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Tailor"},
	[CraftingLib.CodeEnumTradeskill.Runecrafting]	=	{strName = "Runecrafting",	strIcon = ""},
	[CraftingLib.CodeEnumTradeskill.Augmentor]		=	{strName = "Technologist",	strIcon = "IconSprites:Icon_Achievement_UI_Tradeskills_Technologist"},
}
local kTradeskillTiers =
{
	[CraftingLib.CodeEnumTradeskillTier.Zero]			=	"-",
	[CraftingLib.CodeEnumTradeskillTier.Apprentice]		=	"Apprentice",
	[CraftingLib.CodeEnumTradeskillTier.Artisan]		=	"Artisan",
	[CraftingLib.CodeEnumTradeskillTier.Expert]			=	"Expert",
	[CraftingLib.CodeEnumTradeskillTier.GrandMaster]	=	"Grandmaster",
	[CraftingLib.CodeEnumTradeskillTier.Journeyman]		=	"Journeyman",
	[CraftingLib.CodeEnumTradeskillTier.Master]			=	"Master",
	[CraftingLib.CodeEnumTradeskillTier.Novice]			=	"Novice",
}
local kAttributeRootIds =
{
}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function ShortCircuit:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
	self.tSavedConfigurations = {}

    return o
end

function ShortCircuit:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		--"Crafting",
		--"Tradeskills"
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- ShortCircuit OnLoad
-----------------------------------------------------------------------------------------------
function ShortCircuit:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("ShortCircuit.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end


-----------------------------------------------------------------------------------------------
-- ShortCircuit OnDocLoaded
-----------------------------------------------------------------------------------------------
function ShortCircuit:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "ShortCircuitForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.wndMain:Show(false, true)

		-- if the xmlDoc is no longer needed, you should set it to nil
		-- self.xmlDoc = nil
		
		-- Register handlers for events, slash commands and timer, etc.
		-- e.g. Apollo.RegisterEventHandler("KeyDown", "OnKeyDown", self)
		Apollo.RegisterSlashCommand("sc", "OnShortCircuitOn", self)
        Apollo.RegisterEventHandler("GenericEvent_StartCircuitCraft", "OnCircuitOpen", self)
        Apollo.RegisterEventHandler("GenericEvent_InitializeSchematicsTree", "OnInitializeSchematicsTree", self)
        Apollo.RegisterEventHandler("ToggleTradeskills", "OnToggleTradeskills", self)
		Apollo.RegisterEventHandler("GenericError", "OnGenericError", self)
	
		-- Do additional Addon initialization here
		
		-- Initialise a single dialog instance (it'll be re-used) and start it hidden.
		self.wndDialogStore = Apollo.LoadForm(self.xmlDoc, "StoreCircuitForm", wndPreview, self)
		self.wndDialogStore:Show(false)
		-- Initialise a single yes/no dialog instance (it'll be re-used) and start it hidden.
		self.wndDialogYesNo = Apollo.LoadForm(self.xmlDoc, "YesNoDialog", wndPreview, self)
		self.wndDialogYesNo:Show(false)
		
		-- Initialise a single duplicates dialog instance (it'll be re-used) and start it hidden.
		self.wndDialogDuplicate = Apollo.LoadForm(self.xmlDoc, "DuplicateChipsForm", wndPreview, self)
		self.wndDialogDuplicate:Show(false)
		
		-- Create a timer that should fire on crafting completion.
		Apollo.RegisterTimerHandler("ShortCircuit_CraftCompleteTimer", "OnShortCircuit_CraftCompleteTimer", self)
		Apollo.CreateTimer("ShortCircuit_CraftCompleteTimer", 3.25, false)
		Apollo.StopTimer("ShortCircuit_CraftCompleteTimer")
		Apollo.RegisterTimerHandler("ShortCircuit_CraftStartTimer", "OnShortCircuit_CraftStartTimer", self)
		Apollo.CreateTimer("ShortCircuit_CraftStartTimer", 0.25, false)
		Apollo.StopTimer("ShortCircuit_CraftStartTimer")
		
		if not self.strSort then
			self:SetSort("Schematic", self.wndMain:FindChild("Headers"):FindChild("Schematic"))
		end
	
		-- This seems to solve a weird double height scrollbar bug (doesn't work for the variants one though).
		self:PopulateGrid()
		--self:PopulateVariantsGrid()		

		local wndSearchClear = self.wndMain:FindChild("SearchClear")
		wndSearchClear:Show(false)
	end
end

-----------------------------------------------------------------------------------------------
-- ShortCircuit Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

function ShortCircuit:OnShortCircuitOn()
	if not self.strSort then
		self:SetSort("Schematic", self.wndMain:FindChild("Headers"):FindChild("Schematic"))
	end

	self.wndDialogStore:Show(false)
	self.wndDialogYesNo:Show(false)
	self.wndDialogDuplicate:Show(false)
	self.wndMain:Enable(true)
	
	self:PopulateGrid()
	self.wndMain:Invoke() -- show the window
	if self.timerCraftingStationCheck ~= nil then
		self.timerCraftingStationCheck:Stop()
		self.timerCraftingStationCheck = nil
	end	
	
	self.timerCraftingStationCheck = ApolloTimer.Create(1.0, true, "OnCraftingStationCheckTick", self)
end

function ShortCircuit:OnCraftingStationCheckTick()
	if self.wndMain:IsVisible() == false then
		self.timerCraftingStationCheck:Stop()
		self.timerCraftingStationCheck = nil
		return
	end
	
	local bIsAtCraftingStation = CraftingLib.IsAtCraftingStation()
	if bIsAtCraftingStation ~= self.bLastAtCraftingStationStatus then
		self.bLastAtCraftingStationStatus = bIsAtCraftingStation
		self:GridStateUpdate()
	end
end

function ShortCircuit:OnCircuitOpen()
	self:AttachCircuitBoardAdditions()
end

function ShortCircuit:OnInitializeSchematicsTree()
	self:AttachTradeskillAdditions()
end

function ShortCircuit:OnToggleTradeskills()
	self:AttachTradeskillAdditions()
end

function ShortCircuit:OnGenericError(eError, strMessage)
	Print(strMessage)
end

function ShortCircuit:AttachCircuitBoardAdditions()
	local luaCaller = Apollo.GetAddon("Crafting")
	local wndParent = luaCaller.wndMain:FindChild("CraftingForm")
	local wndAdditions = wndParent:FindChild("CircuitBoardAdditions")
	
	if wndAdditions ~= nil then
		wndAdditions:Destroy()
		wndAdditions = nil
	end

	if not wndAdditions then	
		wndAdditions = Apollo.LoadForm(self.xmlDoc, "CircuitBoardAdditions", wndParent , self)	
	end
	
	wndAdditions:ToFront()
end

function ShortCircuit:AttachTradeskillAdditions()
	local wndParent = Apollo.FindWindowByName("TradeskillContainerForm")
	local wndAdditions = wndParent:FindChild("TradeskillAdditions")
	
	if wndAdditions ~= nil then
		wndAdditions:Destroy()
		wndAdditions = nil
	end
	
	if not wndAdditions then	
		wndAdditions = Apollo.LoadForm(self.xmlDoc, "TradeskillAdditions", wndParent, self)	
	end
	
	wndAdditions:ToFront()
end

-----------------------------------------------------------------------------------------------
-- ShortCircuitForm Functions
-----------------------------------------------------------------------------------------------
-- when the OK button is clicked
function ShortCircuit:OnCraft()
	-- Attempt to craft the selected schematic.
	local tItem = self.tSelectedCraft --self.wndSelectedGridItem:GetData()
	
	assert(tItem ~= nil, "No craft selected while attempting to craft.")
	
	if CraftingLib.GetCurrentCraft() == nil then
		CraftingLib.CraftItem(tItem.nSchematicId)
	end
	
	-- Attempt to detect duplicate microchips.
	--[[
	local tLockedChips = {}
	local tStoredChips = {}
	local tCurrentCraft = CraftingLib.GetCurrentCraft()
	for iIndex, tMicrochip in pairs(tCurrentCraft.tMicrochips) do
		if tMicrochip.itemChip then
			local tItem = tMicrochip.itemChip:GetDetailedInfo()
			table.insert(tLockedChips, tItem)
		end
	end

	for iIndex, tMicrochip in pairs(tItem.tMicrochips) do
		local oItem = Item.GetDataFromId(tMicrochip)
		if oItem then
			local tItem = oItem :GetDetailedInfo()
			table.insert(tStoredChips, tItem)
		end
	end

	for iIndex, tLockedChip in pairs(tLockedChips) do
		for iIndex, tStoredChip in pairs(tStoredChips) do
			if tLockedChip.tPrimary.tChipInfo.eProperty ~= nil and tLockedChip.tPrimary.tChipInfo.eProperty == tStoredChip.tPrimary.tChipInfo.eProperty then
				bDuplicateDetected = true
			end
		end
	end
	
	]]--
	
	-- Raise a timer to do the actual craft. Hopefully this will resolve a nil result from GetCurrentCraft 
	-- for some users which I can only presume was a race condition.
	Apollo.StartTimer("ShortCircuit_CraftStartTimer")
end

function ShortCircuit:OnShortCircuit_CraftStartTimer()
	Apollo.StopTimer("ShortCircuit_CraftStartTimer")
	local tItem = self.tSelectedCraft --self.wndSelectedGridItem:GetData()
	
	assert(tItem ~= nil, "No craft selected while attempting to craft.")
		
	local bDuplicateDetected = self:CheckForDuplicateMicrochips(tItem)

	if bDuplicateDetected then
		--Print("Duplicate chips detected, cannot craft.")
		-- Show the resolution dialog.
		self.wndMain:Enable(false)
		self.wndDialogDuplicate:Show(true)
		self.wndDialogDuplicate:ToFront()
	else
		CraftingLib.CompleteCraft(tItem.tMicrochips, tItem.tThresholds)
		Apollo.StartTimer("ShortCircuit_CraftCompleteTimer")
	end	
end

function ShortCircuit:OnShortCircuit_CraftCompleteTimer()
	Print("OnShortCircuit_OnCraftBtnTimer")
	Apollo.StopTimer("ShortCircuit_CraftCompleteTimer")
	self:PopulateGrid()
end

-- when the Cancel button is clicked
function ShortCircuit:OnClose()
	self.wndMain:Close() -- hide the window
	self.timerCraftingStationCheck:Stop()
	self.timerCraftingStationCheck = nil
end

-- when the Remove button is clicked
function ShortCircuit:OnRemove( wndHandler, wndControl, eMouseButton )
	self.wndMain:Enable(false)
	self.wndDialogYesNo:Show(true)
	self.wndDialogYesNo:ToFront()
end

-- when the (+) button is clicked to begin saving a configuration
function ShortCircuit:OnBeginSaveConfiguration( wndHandler, wndControl, eMouseButton )
	-- Pop up the save details dialog.
	local luaCaller = Apollo.GetAddon("Crafting")
	local wndPreview = luaCaller.wndMain:FindChild("CraftingForm")
		
	-- Load schematic information.
	-- Check that we're actually crafting something.
	local tSchematic = CraftingLib.GetCurrentCraft()
	if tSchematic == nil then
		return
	end

	-- Check if all chips have been picked by checking if the Craft button is enabled (easier than implementing all the logic ourselves).
	-- TODO: Could do with disabling the button if not.
	local craftButton = wndPreview:FindChild("CraftButton")
	if not craftButton:IsEnabled() then
		return
	end
	
	self.wndDialogStore:Show(true)
	self.wndDialogStore:ToFront()

	local tSchematicInfo = CraftingLib.GetSchematicInfo(tSchematic.nSchematicId)
	
	-- Set the schematic name, and clear the description.
	local wndSchematicName = self.wndDialogStore:FindChild("LblSchematicName")
	local wndDescription = self.wndDialogStore:FindChild("TxtDescription")
	wndSchematicName:SetText(tSchematicInfo.strName)
end

function ShortCircuit:OnStoreConfigurationSave()
	-- Save the info.
	local wndSchematicName = self.wndDialogStore:FindChild("LblSchematicName")
	local wndDescription = self.wndDialogStore:FindChild("TxtDescription")
	local tSchematic = CraftingLib.GetCurrentCraft()

	local tMicrochips, tThresholds = Apollo.GetAddon("Crafting").luaSchematic:HelperGetUserSelection()
	local nSchematicId = tSchematic.nSchematicId
	local strSchematicName = wndSchematicName:GetText()
	local strDescription = wndDescription:GetText()
	
	local tConfiguration = {}
	tConfiguration.nSchematicId = nSchematicId
	tConfiguration.strSchematicName = strSchematicName
	tConfiguration.strDescription = strDescription
	tConfiguration.tMicrochips = tMicrochips
	tConfiguration.tThresholds = tThresholds
	
	table.insert(self.tSavedConfigurations, tConfiguration)
	
	self.wndDialogStore:Show(false)
end

function ShortCircuit:OnStoreConfigurationCancel()
	-- Close the window.
	self.wndDialogStore:Show(false)
end


function ShortCircuit:PopulateGrid()
	--Print("Populating grid")
	
	--if  not self.tSavedConfigurations then
	--	Print("saved configs is nil")
	--else
	--	Print(#self.tSavedConfigurations)
	--end
	
	--print_r(self.tSavedConfigurations)

	local container = self.wndMain:FindChild("ListContainer")
	
	-- Attempt to remember the last selected item.
	local lastSelected = nil
	if self.wndSelectedGridItem then
		lastSelected = self.wndSelectedGridItem:GetData()
	end
	
	-- Clear out anything that exists in the container already.
	container:DestroyChildren()
	self.wndSelectedGridItem = nil
	self.iGridRow = nil
	
	-- TODO: Not happy about hardcoding this but it'll do.
	local itemHeight = 32
	
	-- Create a list containing only those items that match the search input.
	local tMatchedItems = {}
	
	-- Store the uppercase input (LUA has no case-insensitive string matching...)
	local strSearchInput = self.wndMain:FindChild("SearchInput"):GetText():upper()
	
	for iIndex, tCurrent in pairs(self.tSavedConfigurations) do
		-- Check for matches on item name, schematic name and notes.
		local tSchematicInfo = CraftingLib.GetSchematicInfo(tCurrent.nSchematicId)
		local itemCurrent = tSchematicInfo.itemOutput
		
		-- Store some extra info for later use.
		tCurrent.strTradeskill = kTradeskills[tSchematicInfo.eTradeskillId].strName
		tCurrent.eTier = tSchematicInfo.eTier
		tCurrent.nNumberCraftable = self:DetermineNumberCraftable(tCurrent, tSchematicInfo)
		
		if strSearchInput == "" or string.match(tSchematicInfo.strName:upper(), strSearchInput) or string.match(tCurrent.strDescription:upper(), strSearchInput) then
			table.insert(tMatchedItems, tCurrent)
		end
	end
	
	-- Sort the list of matched items according to the current sort column.
	local sorts = nil
	
	if self.bReverseSort then
		sorts =
		{
			[""] = nil,
			["Tradeskill"] 		= function(a, b) 	return a.strTradeskill < b.strTradeskill		end,
			["Tier"] 			= function(a, b) 	return a.eTier < b.eTier 						end,
			["Item"] 			= function(a, b) 	return a.strSchematicName < b.strSchematicName	end,
			["Schematic"] 		= function(a, b) 	return a.strSchematicName < b.strSchematicName 	end,
			["Details"] 		= function(a, b) 	return a.strTradeskill < b.strTradeskill		end,
			["NumberCraftable"] = function(a, b) 	return a.nNumberCraftable < b.nNumberCraftable 	end,
			["Notes"] 			= function(a, b) 	return a.strDescription < b.strDescription 		end,
		}
	else
		sorts =
		{
			[""] = nil,
			["Tradeskill"] 		= function(a, b) 	return a.strTradeskill > b.strTradeskill		end,
			["Tier"] 			= function(a, b) 	return a.eTier > b.eTier 						end,
			["Item"] 			= function(a, b) 	return a.strSchematicName > b.strSchematicName 	end,
			["Schematic"] 		= function(a, b) 	return a.strSchematicName > b.strSchematicName 	end,
			["Details"] 		= function(a, b) 	return a.strTradeskill > b.strTradeskill		end,
			["NumberCraftable"] = function(a, b) 	return a.nNumberCraftable > b.nNumberCraftable 	end,
			["Notes"] 			= function(a, b) 	return a.strDescription > b.strDescription 		end,
		}
	end
	
	--Print(self.strSort)
	
	table.sort(tMatchedItems, sorts[self.strSort])
	
	for iIndex, tCurrent in pairs(tMatchedItems) do
		local item = Apollo.LoadForm(self.xmlDoc, "ShortCircuitListItem", container, self)
		item:Move(0, itemHeight * (iIndex - 1), container:GetWidth(), itemHeight)
		
		-- Set the details.
		local tSchematicInfo = CraftingLib.GetSchematicInfo(tCurrent.nSchematicId)
		local itemCurrent = tSchematicInfo.itemOutput
		local tItem = itemCurrent:GetDetailedInfo()
		local strIcon = itemCurrent:GetIcon()
		-- Materials count (taken almost directly from CRB addon).
		local nNumCraftable = tCurrent.nNumberCraftable
		
		item:SetData({ tCurrent, nNumCraftable })
		
		-- Columns are:
		-- 1. Trade (icon w/ tooltip)
		-- 2. Tier (Number w/ tooltip)
		-- 3. Icon (Item icon w/ tooltip)
		-- 4. Name (Schematic/item name)
		-- 5. Info (Auto-generated chips listing, images w/ tooltips preferably)
		-- 6. Count (Number)
		-- 7. Notes (Text)
		item:FindChild("ShortCircuitListItemButton"):SetData(iIndex)
		item:FindChild("Tradeskill"):SetSprite(kTradeskills[tSchematicInfo.eTradeskillId].strIcon)
		item:FindChild("Tradeskill"):SetTooltip(kTradeskills[tSchematicInfo.eTradeskillId].strName)
		item:FindChild("Tier"):SetText(tSchematicInfo.eTier)
		item:FindChild("Tier"):SetTooltip(kTradeskillTiers[tSchematicInfo.eTier])
		item:FindChild("ItemIcon"):SetSprite(strIcon)
		Tooltip.GetItemTooltipForm(self, item:FindChild("ItemIcon"), itemCurrent, {bPrimary = true, bSelling = false, itemCompare = itemCurrent:GetEquippedItemForItemType()})
		item:FindChild("SchematicName"):SetText(tCurrent.strSchematicName)
		Tooltip.GetItemTooltipForm(self, item:FindChild("SchematicName"), itemCurrent, {bPrimary = true, bSelling = false, itemCompare = itemCurrent:GetEquippedItemForItemType()})
		item:FindChild("ChipCore"):SetText("")
		item:FindChild("Chip1"):SetText("")
		item:FindChild("Chip2"):SetText("")
		item:FindChild("Chip3"):SetText("")
		item:FindChild("Chip4"):SetText("")
		item:FindChild("Chip5"):SetText("")
		item:FindChild("Chip6"):SetText("")
		item:FindChild("ChipCore"):SetTooltip("")
		item:FindChild("Chip1"):SetTooltip("")
		item:FindChild("Chip2"):SetTooltip("")
		item:FindChild("Chip3"):SetTooltip("")
		item:FindChild("Chip4"):SetTooltip("")
		item:FindChild("Chip5"):SetTooltip("")
		item:FindChild("Chip6"):SetTooltip("")
		item:FindChild("NumberCanCraft"):SetText(nNumCraftable)
		if nNumCraftable == 0 then
			item:FindChild("NumberCanCraft"):SetTextColor("ffff0000")	-- Red
		elseif nNumCraftable < 10 then
			item:FindChild("NumberCanCraft"):SetTextColor("ffffffff")	-- White
		else
			item:FindChild("NumberCanCraft"):SetTextColor("ff00ff00")	-- Green
		end
		item:FindChild("NumberCanCraft"):SetTooltip("Can craft " .. nNumCraftable .. " of this item.")
		item:FindChild("NumberCanCraft"):SetData(nNumCraftable)
		item:FindChild("Notes"):SetText(tCurrent.strDescription)
		item:SetData(tCurrent)
		
		-- Fill in the chips/core details.
		local details = item:FindChild("Details")

		-- Assume first index is the core.
		local nCoreId = tCurrent.tMicrochips[1]
		local itemCore = Item.GetDataFromId(nCoreId)
		if itemCore == nil then
			details:FindChild("ChipCore"):SetText("?")
			details:FindChild("ChipCore"):SetTooltip("No core is selected in this configuration.")
		else
			local itemCoreIcon = itemCore:GetIcon()
			if itemCore ~= nil then
				details:FindChild("ChipCore"):SetText("")
				details:FindChild("ChipCore"):SetSprite(itemCoreIcon)
				Tooltip.GetItemTooltipForm(self, item:FindChild("ChipCore"), itemCore , {bPrimary = true, bSelling = false, itemCompare = itemCurrent:GetEquippedItemForItemType()})
			end
		end

		-- Look at the other array entries and build icons for these also.
		local index = 2
		while index <= #tCurrent.tMicrochips do
			local wndChip = details:FindChild("Chip" .. (index - 1))
			if wndChip ~= nil then
				local itemStat = Item.GetDataFromId(tCurrent.tMicrochips[index])
				if itemStat == nil or tCurrent.tMicrochips[index] == 0 then
					-- Locked chip
					wndChip:SetSprite("IconSprites:Icon_Windows_UI_CRB_Attribute_Random1")
					wndChip:SetTooltip("A random chip will be chosen when crafted.")
					wndChip:SetBGColor("ff808080")
				else
					local itemStatIcon = itemStat:GetIcon()
					local tItemChipInfo = itemStat:GetMicrochipInfo()

					wndChip:SetSprite(itemStatIcon)
					wndChip:SetTooltip(tItemChipInfo.strUnitPropertyName .. ": " .. tItemChipInfo.nBudget)
					wndChip:SetBGColor("white")
					if tItemChipInfo.eType == Item.CodeEnumMicrochipType.Capacitor then
						wndChip:SetBGColor("2x:ffff0000")
					elseif tItemChipInfo.eType == Item.CodeEnumMicrochipType.Inductor then
						wndChip:SetBGColor("ff00ff00")
					end
					wndChip:SetText(tItemChipInfo.nBudget)
				end
			end
			index = index + 1
		end		
	end
	
	container:SetVScrollInfo(#tMatchedItems * itemHeight, container:GetHeight(), 16)

	if lastSelected then
		local wndReselect = container:FindChildByUserData(lastSelected)
		if wndReselect then
			wndReselect:FindChild("ShortCircuitListItemButton"):SetCheck(true)
			self.wndSelectedGridItem = wndReselect
		end
	end
	
	self:GridStateUpdate()
end


function ShortCircuit:GridStateUpdate()
	local blnCanCraft, blnCanRemove, strStatus = self:ValidateGridState()
	self.wndMain:FindChild("LblStatus"):SetText(strStatus)
	self.wndMain:FindChild("btnCraft"):Enable(blnCanCraft)
	self.wndMain:FindChild("btnRemove"):Enable(blnCanRemove)
end

function ShortCircuit:VariantGridStateUpdate()
	local blnCanCraft = self:ValidateVariantGridState()
	self.wndDialogDuplicate:FindChild("UseVariantBtn"):Enable(blnCanCraft)
end

function ShortCircuit:ValidateGridState()
	-- If nothing is selected.
	if self.tSelectedCraft == nil then
		return false, false, "No item is selected to craft."
	end
	
	local tItem = self.tSelectedCraft --self.wndSelectedGridItem:GetData()
	
	-- If bad data is contained in the grid.
	if tItem == nil or tItem.nSchematicId == nil then
		return false, false, "Invalid item data (this is a bug and should never occur)."
	end	
	
	-- If crafting something that isn't the selected item.
	local tCurrentCraft = CraftingLib.GetCurrentCraft()
	if tCurrentCraft ~= nil and tCurrentCraft.nSchematicId ~= tItem.nSchematicId then
		return false, true, "Another item is currently locked into the crafting window. Please finish or abandon it first."
	end
	
	-- If the required materials don't exist.
	if self.wndSelectedGridItem then
		local nNumCraftable = self.wndSelectedGridItem:FindChild("NumberCanCraft"):GetData()
		if nNumCraftable <= 0 then
			return false, true, "You do not have the required materials."
		end
	end

	-- If not near a crafting station.
	if CraftingLib.IsAtCraftingStation() == false then
		return false, true, "You must be near a crafting station to craft."
	end
	
	-- TODO: If the recipe is no longer learned (e.g. switched tradeskills).
	
	return true, true, ""
end

function ShortCircuit:ValidateVariantGridState()
	-- If nothing is selected.
	if self.wndSelectedVariantGridItem == nil then
		return false
	end
	
	local tItem = self.wndSelectedVariantGridItem:GetData()
	
	-- If bad data is contained in the grid.
	if tItem == nil or tItem.nSchematicId == nil then
		return false
	end	
	
	-- If not near a crafting station.
	if CraftingLib.IsAtCraftingStation() == false then
		return false
	end
	
	return true
end


function ShortCircuit:CheckForDuplicateMicrochips(tItem)
	local tLockedChips = {}
	local tStoredChips = {}
	
	local tCurrentCraft = CraftingLib.GetCurrentCraft()
	
	if not tCurrentCraft then
		-- Not crafting, so therefore no duplicates.
		return false
	end
	
	-- No microchips, so therefore no duplicates.
	if not tCurrentCraft.tMicrochips then
		return false
	end
	
	for iIndex, tMicrochip in pairs(tCurrentCraft.tMicrochips) do
		if tMicrochip.itemChip then
			local tItemDetails = tMicrochip.itemChip:GetDetailedInfo()
			table.insert(tLockedChips, tItemDetails )
		end
	end

	for iIndex, tMicrochip in pairs(tItem.tMicrochips) do
		local oItem = Item.GetDataFromId(tMicrochip)
		if oItem then
			local tItem = oItem:GetDetailedInfo()
			table.insert(tStoredChips, tItem)
		end
	end
	
	local bDuplicateDetected = false
	for iIndex, tLockedChip in pairs(tLockedChips) do
		for iIndex, tStoredChip in pairs(tStoredChips) do
			if tLockedChip and tLockedChip.tPrimary and tLockedChip.tPrimary.tChipInfo and tLockedChip.tPrimary.tChipInfo.eProperty then
				if tLockedChip.tPrimary.tChipInfo.eProperty == tStoredChip.tPrimary.tChipInfo.eProperty then
					bDuplicateDetected = true
				end
			end
		end
	end
	
	return bDuplicateDetected
end


function ShortCircuit:PopulateVariantsGrid()
	-- The base schematic is in the data of the selected item.
	local tItem = {}
	if self.tSelectedCraft then
		tItem = self.tSelectedCraft
	end
	
	assert(tItem ~= nil, "Selected craft is nil while populating variants.")
	
	local tSchematicInfo = CraftingLib.GetSchematicInfo(tItem.nSchematicId)
	local tVariants = {}
	for iIndex, tCurrent in pairs(self.tSavedConfigurations) do
		if tCurrent.nSchematicId == tItem.nSchematicId and tCurrent ~= tItem then
			-- Check if the variant is craftable or will have duplicates also.
			local bDuplicateDetected = self:CheckForDuplicateMicrochips(tCurrent)
			local nNumCraftable = self:DetermineNumberCraftable(tCurrent, tSchematicInfo)
			if not bDuplicateDetected and nNumCraftable > 0 then
				table.insert(tVariants, tCurrent)
			end					
		end
	end
	
	-- TODO: Not happy about hardcoding this but it'll do.
	local itemHeight = 32

	local container = self.wndDialogDuplicate:FindChild("ListContainer")
	
	container:DestroyChildren()
	self.wndSelectedVariantGridItem = nil
	self.iVariantGridRow = nil
	
	local wndFirst = nil

	for iIndex, tCurrent in pairs(tVariants) do
		local item = Apollo.LoadForm(self.xmlDoc, "ShortCircuitListItemVariant", container, self)
		item:Move(0, itemHeight * (iIndex - 1), container:GetWidth(), itemHeight)
		
		if not wndFirst then
			wndFirst = item
		end
		
		-- Set the details.
		local tSchematicInfo = CraftingLib.GetSchematicInfo(tCurrent.nSchematicId)
		local itemCurrent = tSchematicInfo.itemOutput
		local tItem = itemCurrent:GetDetailedInfo()
		local strIcon = itemCurrent:GetIcon()

		-- Columns are:
		-- 1. Icon (Item icon w/ tooltip)
		-- 2. Info (Auto-generated chips listing, images w/ tooltips preferably)
		-- 3. Notes (Text)
		item:FindChild("ShortCircuitListItemButton"):SetData(iIndex)
		item:FindChild("ItemIcon"):SetSprite(strIcon)
		Tooltip.GetItemTooltipForm(self, item:FindChild("ItemIcon"), itemCurrent, {bPrimary = true, bSelling = false, itemCompare = itemCurrent:GetEquippedItemForItemType()})
		item:FindChild("ChipCore"):SetText("")
		item:FindChild("Chip1"):SetText("")
		item:FindChild("Chip2"):SetText("")
		item:FindChild("Chip3"):SetText("")
		item:FindChild("Chip4"):SetText("")
		item:FindChild("Chip5"):SetText("")
		item:FindChild("Chip6"):SetText("")
		item:FindChild("ChipCore"):SetTooltip("")
		item:FindChild("Chip1"):SetTooltip("")
		item:FindChild("Chip2"):SetTooltip("")
		item:FindChild("Chip3"):SetTooltip("")
		item:FindChild("Chip4"):SetTooltip("")
		item:FindChild("Chip5"):SetTooltip("")
		item:FindChild("Chip6"):SetTooltip("")
		item:FindChild("Notes"):SetText(tCurrent.strDescription)
		item:SetData(tCurrent)
		
		-- Fill in the chips/core details.
		local details = item:FindChild("Details")

		-- Assume first index is the core.
		local nCoreId = tCurrent.tMicrochips[1]
		local itemCore = Item.GetDataFromId(nCoreId)
		if itemCore == nil then
			details:FindChild("ChipCore"):SetText("?")
			details:FindChild("ChipCore"):SetTooltip("No core is selected in this configuration.")
		else
			local itemCoreIcon = itemCore:GetIcon()
			if itemCore ~= nil then
				details:FindChild("ChipCore"):SetText("")
				details:FindChild("ChipCore"):SetSprite(itemCoreIcon)
				Tooltip.GetItemTooltipForm(self, item:FindChild("ChipCore"), itemCore , {bPrimary = true, bSelling = false, itemCompare = itemCurrent:GetEquippedItemForItemType()})
			end
		end

		-- Look at the other array entries and build icons for these also.
		local index = 2
		while index <= #tCurrent.tMicrochips do
			local wndChip = details:FindChild("Chip" .. (index - 1))
			if wndChip ~= nil then
				local itemStat = Item.GetDataFromId(tCurrent.tMicrochips[index])
				if itemStat == nil or tCurrent.tMicrochips[index] == 0 then
					-- Locked chip
					wndChip:SetSprite("IconSprites:Icon_Windows_UI_CRB_Attribute_Random1")
					wndChip:SetTooltip("A random chip will be chosen when crafted.")
					wndChip:SetBGColor("ff808080")
				else
					local itemStatIcon = itemStat:GetIcon()
					local tItemChipInfo = itemStat:GetMicrochipInfo()

					wndChip:SetSprite(itemStatIcon)
					wndChip:SetTooltip(tItemChipInfo.strUnitPropertyName .. ": " .. tItemChipInfo.nBudget)
					wndChip:SetBGColor("white")
					if tItemChipInfo.eType == Item.CodeEnumMicrochipType.Capacitor then
						wndChip:SetBGColor("2x:ffff0000")
					elseif tItemChipInfo.eType == Item.CodeEnumMicrochipType.Inductor then
						wndChip:SetBGColor("ff00ff00")
					end
					wndChip:SetText(tItemChipInfo.nBudget)
				end
			end
			index = index + 1
		end		
	end
	
	-- For some reason the height is always doubled.
	--Print(#tVariants)
	container:SetVScrollInfo(#tVariants * itemHeight, container:GetHeight(), 16)
	
	-- Select the first item in the list.
	if wndFirst then
		wndFirst:FindChild("ShortCircuitListItemButton"):SetCheck(true)
		self.wndSelectedVariantGridItem = wndFirst
		self.iVariantGridRow = 1
	end
	
	self:VariantGridStateUpdate()
end

function ShortCircuit:DetermineNumberCraftable(tSchematic, tSchematicInfo)
	-- Materials
	local bHaveEnoughMats = true
	local nNumCraftable = 9000
	for key, tMaterial in pairs(tSchematicInfo.tMaterials) do
		if tMaterial.nAmount > 0 then
			local nBackpackCount = tMaterial.itemMaterial:GetBackpackCount()

			nNumCraftable = math.min(nNumCraftable, math.floor(nBackpackCount / tMaterial.nAmount))
			bHaveEnoughMats = bHaveEnoughMats and nBackpackCount >= tMaterial.nAmount
		end
	end

	-- Fake Material (Power Cores)
	if not self.bCoordCraft then
		local tAvailableCores = CraftingLib.GetAvailablePowerCores(tSchematic.nSchematicId)
		if tAvailableCores then -- Some crafts won't have power cores
			local nBackpackCount = 0
			for idx, itemMaterial in pairs(tAvailableCores) do
				nBackpackCount = nBackpackCount + itemMaterial:GetStackCount()
			end
			nNumCraftable = math.min(nNumCraftable, nBackpackCount)
		end
	end
	
	-- Check the core stored in the configuration also.
	local nCoreId = tSchematic.tMicrochips[1]
	local oCore = Item.GetDataFromId(nCoreId)
	nNumCraftable = math.min(nNumCraftable, oCore:GetBackpackCount())
	
	return nNumCraftable
end

function ShortCircuit:OnListContainerSizeChanged( wndHandler, wndControl )
	-- Set the width of all children to the width of the parent.
	local iWidth = wndHandler:GetWidth()
	for key, wndChild in pairs(wndHandler:GetChildren()) do
		--wndChild:SetAnchorOffsets(0, 0, 0, 32)
	end	
end

function ShortCircuit:OnSearchInputTextChanged( wndHandler, wndControl, strText )
	-- TODO: Use a timer here that resets on every keypress, and only fires after 1 second of no extra keypresses.
	
	-- If the text box is empty, then show the watermark, otherwise hide it. Also show/hide the go/clear buttons.
	local wndSearch = self.wndMain:FindChild("Watermark")
	local wndSearchClear = self.wndMain:FindChild("SearchClear")
	if strText == "" then
		wndSearch:Show(true)
		wndSearchClear:Show(false)
	else
		wndSearch:Show(false)
		wndSearchClear:Show(true)
	end
	self:PopulateGrid()
end

function ShortCircuit:OnSearchClear( wndHandler, wndControl, eMouseButton )
	self.wndMain:FindChild("SearchInput"):SetText("")
	local wndSearch = self.wndMain:FindChild("Watermark")
	wndSearch:Show(true)
	self:OnSearchInputTextChanged( wndHandler, wndControl, "" )
end

function ShortCircuit:OnSortByTradeskill( wndHandler, wndControl, eMouseButton )
	self:SetSort("Tradeskill", wndHandler)
	self:PopulateGrid()
end

function ShortCircuit:OnSortByTier( wndHandler, wndControl, eMouseButton )
	self:SetSort("Tier", wndHandler)
	self:PopulateGrid()
end

function ShortCircuit:OnSortByItem( wndHandler, wndControl, eMouseButton )
	self:SetSort("Item", wndHandler)
	self:PopulateGrid()
end

function ShortCircuit:OnSortBySchematic( wndHandler, wndControl, eMouseButton )
	self:SetSort("Schematic", wndHandler)
	self:PopulateGrid()
end

function ShortCircuit:OnSortByDetails( wndHandler, wndControl, eMouseButton )
	self:SetSort("Details", wndHandler)
	self:PopulateGrid()
end

function ShortCircuit:OnSortByNumberCraftable( wndHandler, wndControl, eMouseButton )
	self:SetSort("NumberCraftable", wndHandler)
	self:PopulateGrid()
end

function ShortCircuit:OnSortByNotes( wndHandler, wndControl, eMouseButton )
	self:SetSort("Notes", wndHandler)
	self:PopulateGrid()
end

function ShortCircuit:SetSort(strType, wndHandler)
	if self.wndLastSort then
		self.wndLastSort:ChangeArt("CRB_Basekit:kitBtn_List_MetalSort")
		self.wndLastSort:SetCheck(false)
	end

	if self.strSort == strType then
		self.bReverseSort = not self.bReverseSort
	else
		self.strSort = strType
		self.bReverseSort = true
	end

	self.wndLastSort = wndHandler
	self.wndLastSort:ChangeArt("CRB_Basekit:kitBtn_List_MetalSortArrow")
	if self.bReverseSort then
		self.wndLastSort:SetCheck(true)
	else
		self.wndLastSort:SetCheck(false)
	end
end


---------------------------------------------------------------------------------------------------
-- YesNoDialog Functions
---------------------------------------------------------------------------------------------------

function ShortCircuit:OnYesNoDialogYes( wndHandler, wndControl, eMouseButton )
	-- Find the index of the selected item
	local iRemoveIndex = nil
	for iIndex, tCurrent in pairs(self.tSavedConfigurations) do
		if tCurrent == self.tSelectedCraft then
			iRemoveIndex = iIndex
		end
	end
	
	if iRemoveIndex then
		table.remove(self.tSavedConfigurations, iRemoveIndex)
		self.wndSelectedGridItem = nil
		self.tSelectedCraft = nil
		self.iGridRow = nil
	end
	self:PopulateGrid()
	
	self.wndMain:Enable(true)
	self.wndDialogYesNo:Show(false)
end

function ShortCircuit:OnYesNoDialogNo( wndHandler, wndControl, eMouseButton )
	self.wndMain:Enable(true)
	self.wndDialogYesNo:Show(false)
end


---------------------------------------------------------------------------------------------------
-- ShortCircuitListItem Functions
---------------------------------------------------------------------------------------------------

function ShortCircuit:OnListItemCheck( wndHandler, wndControl, eMouseButton )
	self.iGridRow = wndHandler:GetData()
	self.wndSelectedGridItem = wndHandler:GetParent()
	self.tSelectedCraft = self.wndSelectedGridItem:GetData()
	self:GridStateUpdate()
end

function ShortCircuit:OnListItemUncheck( wndHandler, wndControl, eMouseButton )
	self.iGridRow = nil
	self.wndSelectedGridItem = nil
	self.tSelectedCraft = nil
	self:GridStateUpdate()
end

---------------------------------------------------------------------------------------------------
-- ShortCircuitListItemVariant Functions
---------------------------------------------------------------------------------------------------

function ShortCircuit:OnVariantListItemCheck( wndHandler, wndControl, eMouseButton )
	self.iVariantGridRow = wndHandler:GetData()
	self.wndSelectedVariantGridItem = wndHandler:GetParent()
	self:VariantGridStateUpdate()
end

function ShortCircuit:OnVariantListItemUncheck( wndHandler, wndControl, eMouseButton )
	self.iVariantGridRow = nil
	self.wndSelectedVariantGridItem = nil
	self:VariantGridStateUpdate()
end

---------------------------------------------------------------------------------------------------
-- DuplicateChipsForm Functions
---------------------------------------------------------------------------------------------------

function ShortCircuit:OnCraftUsingVariant( wndHandler, wndControl, eMouseButton )
	self.tSelectedCraft = self.wndSelectedVariantGridItem:GetData()
	self.wndDialogDuplicate:Show(false)
	self.wndMain:Enable(true)
	self:OnCraft()
end

function ShortCircuit:OnCraftManually( wndHandler, wndControl, eMouseButton )
	self.wndDialogDuplicate:Show(false)
	self.wndMain:Enable(true)
	local tItem = self.wndSelectedGridItem:GetData()
	Event_FireGenericEvent("GenericEvent_CraftFromPL", tItem.nSchematicId)
	Event_FireGenericEvent("AlwaysHideTradeskills")
end

function ShortCircuit:OnDuplicateChipsFormShow( wndHandler, wndControl )
	self:PopulateVariantsGrid()
end

-----------------------------------------------------------------------------------------------
-- ShortCircuit Instance
-----------------------------------------------------------------------------------------------
local ShortCircuitInst = ShortCircuit:new()
ShortCircuitInst:Init()


-----------------------------------------------------------------------------------------------
-- ShortCircuit Persistence
-----------------------------------------------------------------------------------------------
function ShortCircuit:OnSave(eLevel)
    if eLevel == GameLib.CodeEnumAddonSaveLevel.Character then
		Print("Saving schematic configuration data.")
	    local save = 
		{
			tSavedConfigurations = self.tSavedConfigurations
	    }
    	return save
    end 

    return nil
end

function ShortCircuit:OnRestore(eLevel, tData)
    if eLevel == GameLib.CodeEnumAddonSaveLevel.Character then
		Print("Loading schematic configuration data.")
	    self.tSavedConfigurations = tData.tSavedConfigurations
    end 
end



-----------------------------------------------------------------------------------------------
-- Helpers
-----------------------------------------------------------------------------------------------
function print_r (t, indent, done)
  done = done or {}
  indent = indent or ''
  local nextIndent -- Storage for next indentation value
  for key, value in pairs (t) do
    if type (value) == "table" and not done [value] then
      nextIndent = nextIndent or
          (indent .. string.rep(' ',string.len(tostring (key))+2))
          -- Shortcut conditional allocation
      done [value] = true
      Print (indent .. "[" .. tostring (key) .. "] => Table {");
      Print  (nextIndent .. "{");
      print_r (value, nextIndent .. string.rep(' ',2), done)
      Print  (nextIndent .. "}");
    else
      Print  (indent .. "[" .. tostring (key) .. "] => " .. tostring (value).."")
    end
  end
end